# Copyright 2022 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from abc import abstractmethod
from typing import Literal

import beartype.typing as tp
from flax import nnx
import jax
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import (
    Float,
    Num,
)

from gpjax.dataset import Dataset
from gpjax.distributions import GaussianDistribution
from gpjax.kernels import RFF
from gpjax.kernels.base import AbstractKernel
from gpjax.likelihoods import (
    AbstractHeteroscedasticLikelihood,
    AbstractLikelihood,
    Gaussian,
    HeteroscedasticGaussian,
    NonGaussian,
)
from gpjax.linalg import (
    Dense,
    Diagonal,
    psd,
    solve,
)
from gpjax.linalg.operations import (
    lower_cholesky,
)
from gpjax.linalg.utils import add_jitter
from gpjax.mean_functions import AbstractMeanFunction
from gpjax.parameters import (
    Parameter,
    Real,
)
from gpjax.typing import (
    Array,
    FunctionalSample,
    KeyArray,
)

K = tp.TypeVar("K", bound=AbstractKernel)
M = tp.TypeVar("M", bound=AbstractMeanFunction)
L = tp.TypeVar("L", bound=AbstractLikelihood)
NGL = tp.TypeVar("NGL", bound=NonGaussian)
GL = tp.TypeVar("GL", bound=Gaussian)
HL = tp.TypeVar("HL", bound=AbstractHeteroscedasticLikelihood)


class AbstractPrior(nnx.Module, tp.Generic[M, K]):
    r"""Abstract Gaussian process prior."""

    def __init__(
        self,
        kernel: K,
        mean_function: M,
        jitter: float = 1e-6,
    ):
        r"""Construct a Gaussian process prior.

        Args:
            kernel: kernel object inheriting from AbstractKernel.
            mean_function: mean function object inheriting from AbstractMeanFunction.
        """
        self.kernel = kernel
        self.mean_function = mean_function
        self.jitter = jitter

    def __call__(
        self,
        test_inputs: Num[Array, "N D"],
        *,
        return_covariance_type: Literal["dense", "diagonal"] = "dense",
    ) -> GaussianDistribution:
        r"""Evaluate the Gaussian process at the given points.

        The output of this function is a ``GaussianDistribution`` from which
        the latent function's mean and covariance can be evaluated and the
        distribution can be sampled.

        Under the hood, ``__call__`` invokes the ``predict`` method. Classes
        inheriting ``AbstractPrior`` should not overwrite ``__call__`` and
        should instead define a ``predict`` method.

        Args:
            test_inputs: Input locations where the GP should be evaluated.
            return_covariance_type: Literal denoting whether to return the full covariance
                of the joint predictive distribution at the test_inputs (dense)
                or just the the standard-deviation of the predictive distribution at
                the test_inputs.

        Returns:
            GaussianDistribution: A multivariate normal random variable representation
                of the Gaussian process.
        """
        return self.predict(
            test_inputs,
            return_covariance_type=return_covariance_type,
        )

    @abstractmethod
    def predict(
        self,
        test_inputs: Num[Array, "N D"],
        *,
        return_covariance_type: Literal["dense", "diagonal"] = "dense",
    ) -> GaussianDistribution:
        r"""Evaluate the predictive distribution.

        Compute the latent function's multivariate normal distribution for a
        given set of parameters. For any class inheriting the `AbstractPrior` class,
        this method must be implemented.

        Args:
            test_inputs: Input locations where the GP should be evaluated.
            return_covariance_type: Literal denoting whether to return the full covariance
                of the joint predictive distribution at the test_inputs (dense)
                or just the the standard-deviation of the predictive distribution at
                the test_inputs.

        Returns:
            GaussianDistribution: A multivariate normal random variable representation
                of the Gaussian process.
        """
        raise NotImplementedError


#######################
# GP Priors
#######################
class Prior(AbstractPrior[M, K]):
    r"""A Gaussian process prior object.

    The GP is parameterised by a
    [mean](https://docs.thomaspinder.com/api/mean_functions/)
    and [kernel](https://docs.thomaspinder.com/api/kernels/base/)
    function.

    A Gaussian process prior parameterised by a mean function $m(\cdot)$ and a kernel
    function $k(\cdot, \cdot)$ is given by
    $p(f(\cdot)) = \mathcal{GP}(m(\cdot), k(\cdot, \cdot))$.

    To invoke a `Prior` distribution, a kernel and mean function must be specified.

    Example:
    ```python
        >>> import gpjax as gpx
        >>> kernel = gpx.kernels.RBF()
        >>> meanf = gpx.mean_functions.Zero()
        >>> prior = gpx.gps.Prior(mean_function=meanf, kernel = kernel)
    ```
    """

    if tp.TYPE_CHECKING:

        @tp.overload
        def __mul__(self, other: GL) -> "ConjugatePosterior[Prior[M, K], GL]": ...

        @tp.overload
        def __mul__(self, other: NGL) -> "NonConjugatePosterior[Prior[M, K], NGL]": ...

        @tp.overload
        def __mul__(self, other: L) -> "AbstractPosterior[Prior[M, K], L]": ...

    def __mul__(self, other):
        r"""Combine the prior with a likelihood to form a posterior distribution.

        The product of a prior and likelihood is proportional to the posterior
        distribution. By computing the product of a GP prior and a likelihood
        object, a posterior GP object will be returned. Mathematically, this can
        be described by:
        ```math
        p(f(\cdot) \mid y) \propto p(y \mid f(\cdot))p(f(\cdot)),
        ```
        where $p(y | f(\cdot))$ is the likelihood and $p(f(\cdot))$ is the prior.

        Example:
        ```pycon
            >>> import gpjax as gpx
            >>> meanf = gpx.mean_functions.Zero()
            >>> kernel = gpx.kernels.RBF()
            >>> prior = gpx.gps.Prior(mean_function=meanf, kernel = kernel)
            >>> likelihood = gpx.likelihoods.Gaussian(num_datapoints=100)
            >>> prior * likelihood
        ```
        Args:
            other (Likelihood): The likelihood distribution of the observed dataset.

        Returns
            Posterior: The relevant GP posterior for the given prior and
                likelihood. Special cases are accounted for where the model
                is conjugate.
        """
        return construct_posterior(prior=self, likelihood=other)

    if tp.TYPE_CHECKING:

        @tp.overload
        def __rmul__(self, other: GL) -> "ConjugatePosterior[Prior[M, K], GL]": ...

        @tp.overload
        def __rmul__(self, other: NGL) -> "NonConjugatePosterior[Prior[M, K], NGL]": ...

        @tp.overload
        def __rmul__(self, other: L) -> "AbstractPosterior[Prior[M, K], L]": ...

    def __rmul__(self, other):
        r"""Combine the prior with a likelihood to form a posterior distribution.

        Reimplement the multiplication operator to allow for order-invariant
        product of a likelihood and a prior i.e., likelihood * prior.

        Args:
            other (Likelihood): The likelihood distribution of the observed
                dataset.

        Returns
            Posterior: The relevant GP posterior for the given prior and
                likelihood. Special cases are accounted for where the model
                is conjugate.
        """
        return self.__mul__(other)

    def predict(
        self,
        test_inputs: Num[Array, "N D"],
        *,
        return_covariance_type: Literal["dense", "diagonal"] = "dense",
    ) -> GaussianDistribution:
        r"""Compute the predictive prior distribution for a given set of
        parameters. The output of this function is a ``GaussianDistribution``
        for a given set of inputs.

        In the following example, we compute the predictive prior distribution
        and then evaluate it on the interval :math:`[0, 1]`:

        Example:
        ```pycon
            >>> import gpjax as gpx
            >>> import jax.numpy as jnp
            >>> kernel = gpx.kernels.RBF()
            >>> mean_function = gpx.mean_functions.Zero()
            >>> prior = gpx.gps.Prior(mean_function=mean_function, kernel=kernel)
            >>> prior.predict(jnp.linspace(0, 1, 100)[:, None])
        ```

        Args:
            test_inputs (Float[Array, "N D"]): The inputs at which to evaluate the
                prior distribution.
            return_covariance_type: Literal denoting whether to return the full covariance
                of the joint predictive distribution at the test_inputs (dense)
                or just the the standard-deviation of the predictive distribution at
                the test_inputs.

        Returns:
            GaussianDistribution: A multivariate normal random variable representation
                of the Gaussian process.
        """

        def _return_full_covariance(
            t: Num[Array, "N D"],
        ) -> Dense:
            Kxx = self.kernel.gram(t)
            Kxx_dense = add_jitter(Kxx.to_dense(), self.jitter)
            Kxx = psd(Dense(Kxx_dense))
            return Kxx

        def _return_diagonal_covariance(
            t: Num[Array, "N D"],
        ) -> Dense:
            Kxx = self.kernel.diagonal(t).diagonal
            Kxx += self.jitter
            Kxx = psd(Dense(Diagonal(Kxx).to_dense()))
            return Kxx

        mean_at_test = self.mean_function(test_inputs)
        cov = jax.lax.cond(
            return_covariance_type == "dense",
            _return_full_covariance,
            _return_diagonal_covariance,
            test_inputs,
        )

        return GaussianDistribution(
            loc=jnp.atleast_1d(mean_at_test.squeeze()), scale=cov
        )

    def sample_approx(
        self,
        num_samples: int,
        key: KeyArray,
        num_features: tp.Optional[int] = 100,
    ) -> FunctionalSample:
        r"""Approximate samples from the Gaussian process prior.

        Build an approximate sample from the Gaussian process prior. This method
        provides a function that returns the evaluations of a sample across any
        given inputs.

        In particular, we approximate the Gaussian processes' prior as the
        finite feature approximation
        $\hat{f}(x) = \sum_{i=1}^m\phi_i(x)\theta_i$ where $\phi_i$ are $m$ features
        sampled from the Fourier feature decomposition of the model's kernel and
        $\theta_i$ are samples from a unit Gaussian.

        A key property of such functional samples is that the same sample draw is
        evaluated for all queries. Consistency is a property that is prohibitively costly
        to ensure when sampling exactly from the GP prior, as the cost of exact sampling
        scales cubically with the size of the sample. In contrast, finite feature representations
        can be evaluated with constant cost regardless of the required number of queries.

        In the following example, we build 10 such samples and then evaluate them
        over the interval $[0, 1]$:

        For a `prior` distribution, the following code snippet will
        build and evaluate an approximate sample.

        Example:
        ```pycon
            >>> import gpjax as gpx
            >>> import jax.numpy as jnp
            >>> import jax.random as jr
            >>> key = jr.key(123)
            >>>
            >>> meanf = gpx.mean_functions.Zero()
            >>> kernel = gpx.kernels.RBF(n_dims=1)
            >>> prior = gpx.gps.Prior(mean_function=meanf, kernel = kernel)
            >>>
            >>> sample_fn = prior.sample_approx(10, key)
            >>> sample_fn(jnp.linspace(0, 1, 100).reshape(-1, 1))
        ```

        Args:
            num_samples (int): The desired number of samples.
            key (KeyArray): The random seed used for the sample(s).
            num_features (int): The number of features used when approximating the
                kernel.

        Returns:
            FunctionalSample: A function representing an approximate sample from the
                Gaussian process prior.
        """

        if (not isinstance(num_samples, int)) or num_samples <= 0:
            raise ValueError("num_samples must be a positive integer")

        fourier_feature_fn = _build_fourier_features_fn(self, num_features, key)

        feature_weights = jr.normal(key, [num_samples, 2 * num_features])

        def sample_fn(test_inputs: Float[Array, "N D"]) -> Float[Array, "N B"]:
            feature_evals = fourier_feature_fn(test_inputs)
            evaluated_sample = jnp.inner(feature_evals, feature_weights)
            return self.mean_function(test_inputs) + evaluated_sample

        return sample_fn


P = tp.TypeVar("P", bound=AbstractPrior)


#######################
# GP Posteriors
#######################from gpjax.linalg.operators import LinearOperator
class AbstractPosterior(nnx.Module, tp.Generic[P, L]):
    r"""Abstract Gaussian process posterior.

    The base GP posterior object conditioned on an observed dataset. All
    posterior objects should inherit from this class.
    """

    def __init__(
        self,
        prior: AbstractPrior[M, K],
        likelihood: L,
        jitter: float = 1e-6,
    ):
        r"""Construct a Gaussian process posterior.

        Args:
            prior (AbstractPrior): The prior distribution.
            likelihood (AbstractLikelihood): The likelihood distribution.
            jitter (float): A small constant added to the diagonal of the
                covariance matrix to ensure numerical stability.
        """
        self.prior = prior
        self.likelihood = likelihood
        self.jitter = jitter

    def __call__(
        self,
        test_inputs: Num[Array, "N D"],
        train_data: Dataset,
        *,
        return_covariance_type: Literal["dense", "diagonal"] = "dense",
    ) -> GaussianDistribution:
        r"""Evaluate the Gaussian process posterior at the given points.

        The output of this function is a ``GaussianDistribution`` from which
        the latent function's mean and covariance can be evaluated and the
        distribution can be sampled.

        Under the hood, ``__call__`` invokes the ``predict`` method. Classes
        inheriting ``AbstractPosterior`` should not overwrite ``__call__`` and
        should instead define a ``predict`` method.

        Args:
            test_inputs: Input locations where the GP should be evaluated.
            train_data: Training dataset to condition on.
            return_covariance_type: Literal denoting whether to return the full covariance
                of the joint predictive distribution at the test_inputs (dense)
                or just the the standard-deviation of the predictive distribution at
                the test_inputs.

        Returns:
            GaussianDistribution: A multivariate normal random variable representation
                of the Gaussian process.
        """
        return self.predict(
            test_inputs,
            train_data,
            return_covariance_type=return_covariance_type,
        )

    @abstractmethod
    def predict(
        self,
        test_inputs: Num[Array, "N D"],
        train_data: Dataset,
        *,
        return_covariance_type: Literal["dense", "diagonal"] = "dense",
    ) -> GaussianDistribution:
        r"""Compute the latent function's multivariate normal distribution for a
        given set of parameters. For any class inheriting the `AbstractPosterior` class,
        this method must be implemented.

        Args:
            test_inputs: Input locations where the GP should be evaluated.
            train_data: Training dataset to condition on.
            return_covariance_type: Literal denoting whether to return the full covariance
                of the joint predictive distribution at the test_inputs (dense)
                or just the the standard-deviation of the predictive distribution at
                the test_inputs.

        Returns:
            GaussianDistribution: A multivariate normal random variable representation
                of the Gaussian process.
        """
        raise NotImplementedError


class LatentPosterior(AbstractPosterior[P, L]):
    r"""A posterior shell used to expose prior structure without inference."""

    def predict(
        self,
        test_inputs: Num[Array, "N D"],
        train_data: Dataset,
        *,
        return_covariance_type: Literal["dense", "diagonal"] = "dense",
    ) -> GaussianDistribution:
        raise NotImplementedError(
            "LatentPosteriors are a lightweight wrapper for priors and do not "
            "implement predictive distributions. Use a variational family for inference."
        )


class ConjugatePosterior(AbstractPosterior[P, GL]):
    r"""A Conjuate Gaussian process posterior object.

    A Gaussian process posterior distribution when the constituent likelihood
    function is a Gaussian distribution. In such cases, the latent function values
    $f$ can be analytically integrated out of the posterior distribution.
    As such, many computational operations can be simplified; something we make use
    of in this object.

    For a Gaussian process prior $p(\mathbf{f})$ and a Gaussian likelihood
    $p(y | \mathbf{f}) = \mathcal{N}(y\mid \mathbf{f}, \sigma^2))$ where
    $\mathbf{f} = f(\mathbf{x})$, the predictive posterior distribution at
    a set of inputs $\mathbf{x}$ is given by
    ```math
    \begin{align}
    p(\mathbf{f}^{\star}\mid \mathbf{y}) & = \int p(\mathbf{f}^{\star}, \mathbf{f} \mid \mathbf{y})\\
        & =\mathcal{N}(\mathbf{f}^{\star} \boldsymbol{\mu}_{\mid \mathbf{y}}, \boldsymbol{\Sigma}_{\mid \mathbf{y}}
    \end{align}
    ```
    where
    ```math
    \begin{align}
    \boldsymbol{\mu}_{\mid \mathbf{y}} & = k(\mathbf{x}^{\star}, \mathbf{x})\left(k(\mathbf{x}, \mathbf{x}')+\sigma^2\mathbf{I}_n\right)^{-1}\mathbf{y}  \\
    \boldsymbol{\Sigma}_{\mid \mathbf{y}} & =k(\mathbf{x}^{\star}, \mathbf{x}^{\star\prime}) -k(\mathbf{x}^{\star}, \mathbf{x})\left( k(\mathbf{x}, \mathbf{x}') + \sigma^2\mathbf{I}_n \right)^{-1}k(\mathbf{x}, \mathbf{x}^{\star}).
    \end{align}
    ```

    Example:
    ```pycon
        >>> import gpjax as gpx
        >>> import jax.numpy as jnp
        >>>
        >>> prior = gpx.gps.Prior(
                mean_function = gpx.mean_functions.Zero(),
                kernel = gpx.kernels.RBF()
            )
        >>> likelihood = gpx.likelihoods.Gaussian(num_datapoints=100)
        >>>
        >>> posterior = prior * likelihood
    ```
    """

    def predict(
        self,
        test_inputs: Num[Array, "M D"],
        train_data: Dataset,
        *,
        return_covariance_type: Literal["dense", "diagonal"] = "dense",
    ) -> GaussianDistribution:
        r"""Query the predictive posterior distribution.

        Conditional on a training data set, compute the GP's posterior
        predictive distribution for a given set of parameters. The returned function
        can be evaluated at a set of test inputs to compute the corresponding
        predictive density.

        The predictive distribution of a conjugate GP is given by
        $$
        p(\mathbf{f}^{\star}\mid \mathbf{y}) & = \int p(\mathbf{f}^{\star} \mathbf{f} \mid \mathbf{y})\\
        & =\mathcal{N}(\mathbf{f}^{\star} \boldsymbol{\mu}_{\mid \mathbf{y}}, \boldsymbol{\Sigma}_{\mid \mathbf{y}}
        $$
        where
        $$
        \boldsymbol{\mu}_{\mid \mathbf{y}} & = k(\mathbf{x}^{\star}, \mathbf{x})\left(k(\mathbf{x}, \mathbf{x}')+\sigma^2\mathbf{I}_n\right)^{-1}\mathbf{y} \\
        \boldsymbol{\Sigma}_{\mid \mathbf{y}} & =k(\mathbf{x}^{\star}, \mathbf{x}^{\star\prime}) -k(\mathbf{x}^{\star}, \mathbf{x})\left( k(\mathbf{x}, \mathbf{x}') + \sigma^2\mathbf{I}_n \right)^{-1}k(\mathbf{x}, \mathbf{x}^{\star}).
        $$

        The conditioning set is a GPJax `Dataset` object, whilst predictions
        are made on a regular Jax array.

        Example:
        ```pycon
            >>> import gpjax as gpx
            >>> import jax.numpy as jnp
            >>>
            >>> xtrain = jnp.linspace(0, 1).reshape(-1, 1)
            >>> ytrain = jnp.sin(xtrain)
            >>> D = gpx.Dataset(X=xtrain, y=ytrain)
            >>> xtest = jnp.linspace(0, 1).reshape(-1, 1)
            >>>
            >>> prior = gpx.gps.Prior(mean_function = gpx.mean_functions.Zero(), kernel = gpx.kernels.RBF())
            >>> posterior = prior * gpx.likelihoods.Gaussian(num_datapoints = D.n)
            >>> predictive_dist = posterior(xtest, D)
        ```

        Args:
            test_inputs (Num[Array, "N D"]): A Jax array of test inputs at which the
                predictive distribution is evaluated.
            train_data (Dataset): A `gpx.Dataset` object that contains the input and
                output data used for training dataset.
            return_covariance_type: Literal denoting whether to return the full covariance
                of the joint predictive distribution at the test_inputs (dense)
                or just the the standard-deviation of the predictive distribution at
                the test_inputs.

        Returns:
            GaussianDistribution: A function that accepts an input array and
                returns the predictive distribution as a `GaussianDistribution`.
        """
        x = train_data.X
        y = train_data.y
        # Observation noise o²
        obs_noise = jnp.square(self.likelihood.obs_stddev[...])
        mx = self.prior.mean_function(x)
        # Precompute Gram matrix, Kxx, at training inputs, x
        Kxx = self.prior.kernel.gram(x)
        Kxx = add_jitter(Kxx.to_dense(), self.jitter)

        Sigma_dense = Kxx + jnp.eye(Kxx.shape[0]) * obs_noise
        Sigma = psd(Dense(Sigma_dense))
        L_sigma = lower_cholesky(Sigma)

        Kxt = self.prior.kernel.cross_covariance(x, test_inputs)

        L_inv_Kxt = solve(L_sigma, Kxt)
        L_inv_y_diff = solve(L_sigma, y - mx)

        mean_t = self.prior.mean_function(test_inputs)
        mean = mean_t + jnp.matmul(L_inv_Kxt.T, L_inv_y_diff)

        def _return_full_covariance(
            L_inv_Kxt: Num[Array, "N M"],
            t: Num[Array, "M D"],
        ) -> Dense:
            Ktt = self.prior.kernel.gram(t)
            covariance = Ktt.to_dense() - jnp.matmul(L_inv_Kxt.T, L_inv_Kxt)
            covariance = add_jitter(covariance, self.prior.jitter)
            covariance = psd(Dense(covariance))
            return covariance

        def _return_diagonal_covariance(
            L_inv_Kxt: Num[Array, "N M"],
            t: Num[Array, "M D"],
        ) -> Dense:
            Ktt = self.prior.kernel.diagonal(t).diagonal
            covariance = Ktt - jnp.einsum("ij, ji->i", L_inv_Kxt.T, L_inv_Kxt)
            covariance += self.prior.jitter
            covariance = psd(Dense(jnp.diag(jnp.atleast_1d(covariance.squeeze()))))
            return covariance

        cov = jax.lax.cond(
            return_covariance_type == "dense",
            _return_full_covariance,
            _return_diagonal_covariance,
            L_inv_Kxt,
            test_inputs,
        )

        return GaussianDistribution(loc=jnp.atleast_1d(mean.squeeze()), scale=cov)

    def sample_approx(
        self,
        num_samples: int,
        train_data: Dataset,
        key: KeyArray,
        num_features: int | None = 100,
    ) -> FunctionalSample:
        r"""Draw approximate samples from the Gaussian process posterior.

        Build an approximate sample from the Gaussian process posterior. This method
        provides a function that returns the evaluations of a sample across any given
        inputs.

        Unlike when building approximate samples from a Gaussian process prior, decompositions
        based on Fourier features alone rarely give accurate samples. Therefore, we must also
        include an additional set of features (known as canonical features) to better model the
        transition from Gaussian process prior to Gaussian process posterior. For more details
        see [Wilson et. al. (2020)](https://arxiv.org/abs/2002.09309).

        In particular, we approximate the Gaussian processes' posterior as the finite
        feature approximation
        $\hat{f}(x) = \sum_{i=1}^m \phi_i(x)\theta_i + \sum{j=1}^N v_jk(.,x_j)$
        where $\phi_i$ are m features sampled from the Fourier feature decomposition of
        the model's kernel and $k(., x_j)$ are N canonical features. The Fourier
        weights $\theta_i$ are samples from a unit Gaussian. See
        [Wilson et. al. (2020)](https://arxiv.org/abs/2002.09309) for expressions
        for the canonical weights $v_j$.

        A key property of such functional samples is that the same sample draw is
        evaluated for all queries. Consistency is a property that is prohibitively costly
        to ensure when sampling exactly from the GP prior, as the cost of exact sampling
        scales cubically with the size of the sample. In contrast, finite feature representations
        can be evaluated with constant cost regardless of the required number of queries.

        Args:
            num_samples (int): The desired number of samples.
            key (KeyArray): The random seed used for the sample(s).
            num_features (int): The number of features used when approximating the
                kernel.

        Returns:
            FunctionalSample: A function representing an approximate sample from the Gaussian
                process prior.
        """
        if (not isinstance(num_samples, int)) or num_samples <= 0:
            raise ValueError("num_samples must be a positive integer")

        # sample fourier features
        fourier_feature_fn = _build_fourier_features_fn(self.prior, num_features, key)

        fourier_weights = jr.normal(key, [num_samples, 2 * num_features])

        obs_var = self.likelihood.obs_stddev[...] ** 2
        Kxx = self.prior.kernel.gram(train_data.X)
        Sigma = Dense(add_jitter(Kxx.to_dense(), obs_var + self.jitter))
        eps = jnp.sqrt(obs_var) * jr.normal(key, [train_data.n, num_samples])
        y = train_data.y - self.prior.mean_function(train_data.X)
        Phi = fourier_feature_fn(train_data.X)
        canonical_weights = solve(
            Sigma,
            y + eps - jnp.inner(Phi, fourier_weights),
        )  # [N, B]

        def sample_fn(test_inputs: Float[Array, "n D"]) -> Float[Array, "n B"]:
            fourier_features = fourier_feature_fn(test_inputs)
            weight_space_contribution = jnp.inner(fourier_features, fourier_weights)
            canonical_features = self.prior.kernel.cross_covariance(
                test_inputs, train_data.X
            )
            function_space_contribution = jnp.matmul(
                canonical_features, canonical_weights
            )

            return (
                self.prior.mean_function(test_inputs)
                + weight_space_contribution
                + function_space_contribution
            )

        return sample_fn


class NonConjugatePosterior(AbstractPosterior[P, NGL]):
    r"""A non-conjugate Gaussian process posterior object.

    A Gaussian process posterior object for models where the likelihood is
    non-Gaussian. Unlike the `ConjugatePosterior` object, the
    `NonConjugatePosterior` object does not provide an exact marginal
    log-likelihood function. Instead, the `NonConjugatePosterior` object
    represents the posterior distributions as a function of the model's
    hyperparameters and the latent function. Markov chain Monte Carlo,
    variational inference, or Laplace approximations can then be used to sample
    from, or optimise an approximation to, the posterior distribution.
    """

    latent: nnx.Intermediate[Float[Array, "N 1"]]

    def __init__(
        self,
        prior: P,
        likelihood: NGL,
        latent: tp.Union[Float[Array, "N 1"], Parameter, None] = None,
        jitter: float = 1e-6,
        key: KeyArray = jr.key(42),
    ):
        r"""Construct a non-conjugate Gaussian process posterior.

        Args:
            prior (AbstractPrior): The prior distribution.
            likelihood (AbstractLikelihood): The likelihood distribution.
            jitter (float): A small constant added to the diagonal of the
                covariance matrix to ensure numerical stability.
        """
        super().__init__(prior=prior, likelihood=likelihood, jitter=jitter)

        if latent is None:
            latent = jr.normal(key, shape=(self.likelihood.num_datapoints, 1))

        # TODO: static or intermediate?
        self.latent = latent if isinstance(latent, Parameter) else Real(latent)
        self.key = key

    def predict(
        self,
        test_inputs: Num[Array, "M D"],
        train_data: Dataset,
        *,
        return_covariance_type: Literal["dense", "diagonal"] = "dense",
    ) -> GaussianDistribution:
        r"""Query the predictive posterior distribution.

        Conditional on a set of training data, compute the GP's posterior
        predictive distribution for a given set of parameters. The returned
        function can be evaluated at a set of test inputs to compute the
        corresponding predictive density. Note, to gain predictions on the scale
        of the original data, the returned distribution will need to be
        transformed through the likelihood function's inverse link function.

        Args:
        test_inputs (Num[Array, "N D"]): A Jax array of test inputs at which the
            predictive distribution is evaluated.
        train_data (Dataset): A `gpx.Dataset` object that contains the input
            and output data used for training dataset.
        return_covariance_type: Literal denoting whether to return the full covariance
            of the joint predictive distribution at the test_inputs (dense)
            or just the the standard-deviation of the predictive distribution at
            the test_inputs.

        Returns:
            GaussianDistribution: A function that accepts an
                input array and returns the predictive distribution as
                a `dx.Distribution`.
        """
        x = train_data.X
        t = test_inputs
        mean_function = self.prior.mean_function
        kernel = self.prior.kernel

        # Precompute lower triangular of Gram matrix
        Kxx = kernel.gram(x)
        Kxx_dense = add_jitter(Kxx.to_dense(), self.prior.jitter)
        Kxx = psd(Dense(Kxx_dense))
        Lx = lower_cholesky(Kxx)

        Kxt = kernel.cross_covariance(x, t)
        # Lx⁻¹ Kxt
        Lx_inv_Kxt = solve(Lx, Kxt)

        mean_t = mean_function(t)
        # Whitened function values, wx, corresponding to the inputs, x
        wx = self.latent[...]

        # μt + Ktx Lx⁻¹ wx
        mean = mean_t + jnp.matmul(Lx_inv_Kxt.T, wx)

        def _return_full_covariance(
            Lx_inv_Kxt: Num[Array, "N M"],
            t: Num[Array, "M D"],
        ) -> Dense:
            Ktt = kernel.gram(t)
            covariance = Ktt.to_dense() - jnp.matmul(Lx_inv_Kxt.T, Lx_inv_Kxt)
            covariance = add_jitter(covariance, self.prior.jitter)
            covariance = psd(Dense(covariance))

            return covariance

        def _return_diagonal_covariance(
            Lx_inv_Kxt: Num[Array, "N M"],
            t: Num[Array, "M D"],
        ) -> Dense:
            Ktt = kernel.diagonal(t).diagonal
            covariance = Ktt - jnp.einsum("ij, ji->i", Lx_inv_Kxt.T, Lx_inv_Kxt)
            covariance += self.prior.jitter
            # It would be nice to return a Diagonal here, but the pytree needs
            # to be the same for both cond branches and the other branch needs
            # to return a Dense.
            # They are both LinearOperators, but they inherit from that class
            # and hence are not the same pytree anymore.
            covariance = psd(Dense(jnp.diag(jnp.atleast_1d(covariance.squeeze()))))

            return covariance

        cov = jax.lax.cond(
            return_covariance_type == "dense",
            _return_full_covariance,
            _return_diagonal_covariance,
            Lx_inv_Kxt,
            test_inputs,
        )

        return GaussianDistribution(jnp.atleast_1d(mean.squeeze()), cov)


class HeteroscedasticPosterior(LatentPosterior[P, HL]):
    r"""Posterior shell for heteroscedastic likelihoods.

    The posterior retains both the signal and noise priors; inference is delegated
    to variational families and specialised objectives.
    """

    def __init__(
        self,
        prior: AbstractPrior[M, K],
        likelihood: HL,
        jitter: float = 1e-6,
    ):
        if likelihood.noise_prior is None:
            raise ValueError("Heteroscedastic likelihoods require a noise_prior.")
        super().__init__(prior=prior, likelihood=likelihood, jitter=jitter)
        self.noise_prior = likelihood.noise_prior
        self.noise_posterior = LatentPosterior(
            prior=self.noise_prior, likelihood=likelihood, jitter=jitter
        )


class ChainedPosterior(HeteroscedasticPosterior[P, HL]):
    r"""Posterior routed for heteroscedastic likelihoods using chained bounds."""

    def __init__(
        self,
        prior: AbstractPrior[M, K],
        likelihood: HL,
        jitter: float = 1e-6,
    ):
        super().__init__(prior=prior, likelihood=likelihood, jitter=jitter)


#######################
# Utils
#######################


@tp.overload
def construct_posterior(prior: P, likelihood: GL) -> ConjugatePosterior[P, GL]: ...


@tp.overload
def construct_posterior(prior: P, likelihood: NGL) -> NonConjugatePosterior[P, NGL]: ...


@tp.overload
def construct_posterior(
    prior: P, likelihood: HeteroscedasticGaussian
) -> HeteroscedasticPosterior[P, HeteroscedasticGaussian]: ...


@tp.overload
def construct_posterior(
    prior: P, likelihood: AbstractHeteroscedasticLikelihood
) -> ChainedPosterior[P, AbstractHeteroscedasticLikelihood]: ...


def construct_posterior(
    prior: AbstractPrior, likelihood: AbstractLikelihood
) -> "AbstractPosterior":
    r"""Utility function for constructing a posterior object from a prior and
    likelihood. The function will automatically select the correct posterior
    object based on the likelihood.

    Args:
        prior (Prior): The Prior distribution.
        likelihood (AbstractLikelihood): The likelihood that represents our
            beliefs around the distribution of the data.

    Returns
    -------
    AbstractPosterior
        A posterior distribution. If the likelihood is Gaussian, then a
        `ConjugatePosterior` will be returned. Otherwise, a
        `NonConjugatePosterior` will be returned.
    """
    if isinstance(likelihood, Gaussian):
        return ConjugatePosterior(prior=prior, likelihood=likelihood)

    if (
        isinstance(likelihood, HeteroscedasticGaussian)
        and likelihood.supports_tight_bound()
    ):
        return HeteroscedasticPosterior(prior=prior, likelihood=likelihood)

    if isinstance(likelihood, AbstractHeteroscedasticLikelihood):
        return ChainedPosterior(prior=prior, likelihood=likelihood)

    return NonConjugatePosterior(prior=prior, likelihood=likelihood)


def _build_fourier_features_fn(
    prior: Prior, num_features: int, key: KeyArray
) -> tp.Callable[[Float[Array, "N D"]], Float[Array, "N L"]]:
    r"""Return a function that evaluates features sampled from the Fourier feature
    decomposition of the prior's kernel.

    Args:
        prior (Prior): The Prior distribution.
        num_features (int): The number of feature functions to be sampled.
        key (KeyArray): The random seed used.

    Returns
    -------
    Callable
        A callable function evaluating the sampled feature functions.
    """
    if (not isinstance(num_features, int)) or num_features <= 0:
        raise ValueError("num_features must be a positive integer")

    # Approximate kernel with feature decomposition
    approximate_kernel = RFF(
        base_kernel=prior.kernel, num_basis_fns=num_features, key=key
    )

    def eval_fourier_features(test_inputs: Float[Array, "N D"]) -> Float[Array, "N L"]:
        Phi = approximate_kernel.compute_features(x=test_inputs)
        Phi *= jnp.sqrt(prior.kernel.variance[...] / num_features)
        return Phi

    return eval_fourier_features


__all__ = [
    "AbstractPosterior",
    "AbstractPrior",
    "ChainedPosterior",
    "ConjugatePosterior",
    "HeteroscedasticPosterior",
    "LatentPosterior",
    "NonConjugatePosterior",
    "Prior",
    "construct_posterior",
]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import TYPE_CHECKING

import beartype.typing as tp
from flax import nnx
import jax
from jax import vmap
import jax.nn as jnn
import jax.numpy as jnp
import jax.scipy as jsp
from jaxtyping import Float
import numpy as np
import numpyro.distributions as npd

from gpjax.distributions import GaussianDistribution
from gpjax.integrators import (
    AbstractIntegrator,
    AnalyticalGaussianIntegrator,
    GHQuadratureIntegrator,
)
from gpjax.parameters import (
    NonNegativeReal,
)
from gpjax.typing import (
    Array,
    ScalarFloat,
)

if TYPE_CHECKING:
    from gpjax.gps import Prior


@dataclass(slots=True)
class NoiseMoments:
    log_variance: Array
    inv_variance: Array
    variance: Array


jax.tree_util.register_pytree_node(
    NoiseMoments,
    lambda x: ((x.log_variance, x.inv_variance, x.variance), None),
    lambda _, x: NoiseMoments(*x),
)


class AbstractLikelihood(nnx.Module):
    r"""Abstract base class for likelihoods.

    All likelihoods must inherit from this class and implement the `predict` and
    `link_function` methods.
    """

    def __init__(
        self,
        num_datapoints: int,
        integrator: AbstractIntegrator = GHQuadratureIntegrator(),
    ):
        """Initializes the likelihood.

        Args:
            num_datapoints (int): the number of data points.
            integrator (AbstractIntegrator): The integrator to be used for computing expected log
                likelihoods. Must be an instance of `AbstractIntegrator`.
        """
        self.num_datapoints = num_datapoints
        self.integrator = integrator

    def __call__(
        self, dist: tp.Union[npd.MultivariateNormal, GaussianDistribution]
    ) -> npd.Distribution:
        r"""Evaluate the likelihood function at a given predictive distribution.

        Args:
            dist: The predictive distribution to evaluate the likelihood at.

        Returns:
            The predictive distribution.
        """
        return self.predict(dist)

    @abc.abstractmethod
    def predict(
        self, dist: tp.Union[npd.MultivariateNormal, GaussianDistribution]
    ) -> npd.Distribution:
        r"""Evaluate the likelihood function at a given predictive distribution.

        Args:
            dist: The predictive distribution to evaluate the likelihood at.

        Returns:
            npd.Distribution: The predictive distribution.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def link_function(self, f: Float[Array, ...]) -> npd.Distribution:
        r"""Return the link function of the likelihood function.

        Args:
            f (Float[Array, "..."]): the latent Gaussian process values.

        Returns:
            npd.Distribution: The distribution of observations, y, given values of the
                Gaussian process, f.
        """
        raise NotImplementedError

    def expected_log_likelihood(
        self,
        y: Float[Array, "N D"],
        mean: Float[Array, "N D"],
        variance: Float[Array, "N D"],
        mean_g: tp.Optional[Float[Array, "N D"]] = None,
        variance_g: tp.Optional[Float[Array, "N D"]] = None,
        **_: tp.Any,
    ) -> Float[Array, " N"]:
        r"""Compute the expected log likelihood.

        For a variational distribution $q(f)\sim\mathcal{N}(m, s)$ and a likelihood
        $p(y|f)$, compute the expected log likelihood:
        ```math
        \mathbb{E}_{q(f)}\left[\log p(y|f)\right]
        ```

        Args:
            y (Float[Array, 'N D']): The observed response variable.
            mean (Float[Array, 'N D']): The variational mean.
            variance (Float[Array, 'N D']): The variational variance.
            mean_g (Float[Array, 'N D']): Optional moments of the latent noise
                process for heteroscedastic likelihoods.
            variance_g (Float[Array, 'N D']): Optional moments of the latent noise
                process for heteroscedastic likelihoods.
            **_: Unused extra arguments for compatibility with specialised
                likelihoods.

        Returns:
            ScalarFloat: The expected log likelihood.
        """
        log_prob = vmap(lambda f, y: self.link_function(f).log_prob(y))
        return self.integrator(
            fun=log_prob, y=y, mean=mean, variance=variance, likelihood=self
        )


class AbstractNoiseTransform(nnx.Module):
    """Abstract base class for noise transformations."""

    @abc.abstractmethod
    def __call__(self, x: Float[Array, ...]) -> Float[Array, ...]:
        """Transform the input noise signal."""
        raise NotImplementedError

    @abc.abstractmethod
    def moments(
        self, mean: Float[Array, ...], variance: Float[Array, ...]
    ) -> NoiseMoments:
        """Compute the moments of the transformed noise signal."""
        raise NotImplementedError


class LogNormalTransform(AbstractNoiseTransform):
    """Log-normal noise transformation."""

    def __call__(self, x: Float[Array, ...]) -> Float[Array, ...]:
        return jnp.exp(x)

    def moments(
        self, mean: Float[Array, ...], variance: Float[Array, ...]
    ) -> NoiseMoments:
        expected_variance = jnp.exp(mean + 0.5 * variance)
        expected_log_variance = mean
        expected_inv_variance = jnp.exp(-mean + 0.5 * variance)
        return NoiseMoments(
            log_variance=expected_log_variance,
            inv_variance=expected_inv_variance,
            variance=expected_variance,
        )


class SoftplusTransform(AbstractNoiseTransform):
    """Softplus noise transformation."""

    def __init__(self, num_points: int = 20):
        self.num_points = num_points

    def __call__(self, x: Float[Array, ...]) -> Float[Array, ...]:
        return jnn.softplus(x)

    def moments(
        self, mean: Float[Array, ...], variance: Float[Array, ...]
    ) -> NoiseMoments:
        quad_x, quad_w = np.polynomial.hermite.hermgauss(self.num_points)
        quad_w = jnp.asarray(quad_w / jnp.sqrt(jnp.pi))
        quad_x = jnp.asarray(quad_x)

        std = jnp.sqrt(variance)
        samples = mean[..., None] + jnp.sqrt(2.0) * std[..., None] * quad_x
        sigma2 = self(samples)
        log_sigma2 = jnp.log(sigma2)
        inv_sigma2 = 1.0 / sigma2

        expected_variance = jnp.sum(sigma2 * quad_w, axis=-1)
        expected_log_variance = jnp.sum(log_sigma2 * quad_w, axis=-1)
        expected_inv_variance = jnp.sum(inv_sigma2 * quad_w, axis=-1)

        return NoiseMoments(
            log_variance=expected_log_variance,
            inv_variance=expected_inv_variance,
            variance=expected_variance,
        )


class AbstractHeteroscedasticLikelihood(AbstractLikelihood):
    r"""Base class for heteroscedastic likelihoods with latent noise processes."""

    def __init__(
        self,
        num_datapoints: int,
        noise_prior: Prior,
        noise_transform: tp.Union[
            AbstractNoiseTransform,
            tp.Callable[[Float[Array, ...]], Float[Array, ...]],
        ] = SoftplusTransform(),
        integrator: AbstractIntegrator = GHQuadratureIntegrator(),
    ):
        self.noise_prior = noise_prior

        if isinstance(noise_transform, AbstractNoiseTransform):
            self.noise_transform = noise_transform
        else:
            transform_name = getattr(noise_transform, "__name__", "")
            if noise_transform is jnp.exp or transform_name == "exp":
                self.noise_transform = LogNormalTransform()
            else:
                # Default to SoftplusTransform for softplus or unknown callables (legacy behavior used quadrature)
                # Note: If an unknown callable is passed, we technically use SoftplusTransform which applies softplus.
                # Users should implement AbstractNoiseTransform for custom transforms.
                self.noise_transform = SoftplusTransform()

        super().__init__(num_datapoints=num_datapoints, integrator=integrator)

    def __call__(
        self,
        dist: tp.Union[npd.MultivariateNormal, GaussianDistribution],
        noise_dist: tp.Optional[
            tp.Union[npd.MultivariateNormal, GaussianDistribution]
        ] = None,
    ) -> npd.Distribution:
        return self.predict(dist, noise_dist)

    def supports_tight_bound(self) -> bool:
        """Return whether the tighter bound from Lázaro-Gredilla & Titsias (2011)
        is applicable."""
        return False

    def noise_statistics(
        self, mean: Float[Array, "N D"], variance: Float[Array, "N D"]
    ) -> NoiseMoments:
        r"""Moment matching of the transformed noise process.

        Args:
            mean: Mean of the latent noise GP.
            variance: Variance of the latent noise GP.

        Returns:
            NoiseMoments: Expected log variance, inverse variance, and variance.
        """
        return self.noise_transform.moments(mean, variance)

    def expected_log_likelihood(
        self,
        y: Float[Array, "N D"],
        mean: Float[Array, "N D"],
        variance: Float[Array, "N D"],
        mean_g: tp.Optional[Float[Array, "N D"]] = None,
        variance_g: tp.Optional[Float[Array, "N D"]] = None,
        **kwargs: tp.Any,
    ) -> Float[Array, " N"]:
        raise NotImplementedError


class Gaussian(AbstractLikelihood):
    r"""Gaussian likelihood object."""

    def __init__(
        self,
        num_datapoints: int,
        obs_stddev: tp.Union[ScalarFloat, Float[Array, "#N"], NonNegativeReal] = 1.0,
        integrator: AbstractIntegrator = AnalyticalGaussianIntegrator(),
    ):
        r"""Initializes the Gaussian likelihood.

        Args:
            num_datapoints (int): the number of data points.
            obs_stddev (Union[ScalarFloat, Float[Array, "#N"]]): the standard deviation
                of the Gaussian observation noise.
            integrator (AbstractIntegrator): The integrator to be used for computing expected log
                likelihoods. Must be an instance of `AbstractIntegrator`. For the Gaussian likelihood, this defaults to
                the `AnalyticalGaussianIntegrator`, as the expected log likelihood can be computed analytically.
        """
        if not isinstance(obs_stddev, NonNegativeReal):
            obs_stddev = NonNegativeReal(jnp.asarray(obs_stddev))
        self.obs_stddev = obs_stddev

        super().__init__(num_datapoints, integrator)

    def link_function(self, f: Float[Array, ...]) -> npd.Normal:
        r"""The link function of the Gaussian likelihood.

        Args:
            f (Float[Array, "..."]): Function values.

        Returns:
            npd.Normal: The likelihood function.
        """
        return npd.Normal(loc=f, scale=self.obs_stddev[...].astype(f.dtype))

    def predict(
        self, dist: tp.Union[npd.MultivariateNormal, GaussianDistribution]
    ) -> npd.MultivariateNormal:
        r"""Evaluate the Gaussian likelihood.

        Evaluate the Gaussian likelihood function at a given predictive
        distribution. Computationally, this is equivalent to summing the
        observation noise term to the diagonal elements of the predictive
        distribution's covariance matrix.

        Args:
            dist (npd.Distribution): The Gaussian process posterior,
                evaluated at a finite set of test points.

        Returns:
            npd.Distribution: The predictive distribution.
        """
        n_data = dist.event_shape[0]
        cov = dist.covariance_matrix
        noisy_cov = cov.at[jnp.diag_indices(n_data)].add(self.obs_stddev[...] ** 2)

        return npd.MultivariateNormal(dist.mean, noisy_cov)


class HeteroscedasticGaussian(AbstractHeteroscedasticLikelihood):
    def predict(
        self,
        dist: tp.Union[npd.MultivariateNormal, GaussianDistribution],
        noise_dist: tp.Optional[
            tp.Union[npd.MultivariateNormal, GaussianDistribution]
        ] = None,
    ) -> npd.MultivariateNormal:
        if noise_dist is None:
            raise ValueError(
                "noise_dist must be provided for heteroscedastic prediction."
            )

        n_data = dist.event_shape[0]
        noise_mean = noise_dist.mean
        noise_variance = jnp.diag(noise_dist.covariance_matrix)
        noise_stats = self.noise_statistics(
            noise_mean[..., None], noise_variance[..., None]
        )

        cov = dist.covariance_matrix
        noisy_cov = cov.at[jnp.diag_indices(n_data)].add(noise_stats.variance.squeeze())

        return npd.MultivariateNormal(dist.mean, noisy_cov)

    def link_function(self, f: Float[Array, ...]) -> npd.Normal:
        sigma2 = self.noise_transform(jnp.zeros_like(f))
        return npd.Normal(loc=f, scale=jnp.sqrt(sigma2))

    def expected_log_likelihood(
        self,
        y: Float[Array, "N D"],
        mean: Float[Array, "N D"],
        variance: Float[Array, "N D"],
        mean_g: tp.Optional[Float[Array, "N D"]] = None,
        variance_g: tp.Optional[Float[Array, "N D"]] = None,
        noise_stats: tp.Optional[NoiseMoments] = None,
        return_parts: bool = False,
        **_: tp.Any,
    ) -> tp.Union[Float[Array, " N"], tuple[Float[Array, " N"], NoiseMoments]]:
        if mean_g is None or variance_g is None:
            raise ValueError(
                "mean_g and variance_g must be provided for heteroscedastic models."
            )

        if noise_stats is None:
            noise_stats = self.noise_statistics(mean_g, variance_g)
        sq_error = jnp.square(y - mean)
        log2pi = jnp.log(2.0 * jnp.pi)
        expected = -0.5 * (
            log2pi
            + noise_stats.log_variance
            + (sq_error + variance) * noise_stats.inv_variance
        )
        expected_sum = jnp.sum(expected, axis=1)
        if return_parts:
            return expected_sum, noise_stats
        return expected_sum

    def supports_tight_bound(self) -> bool:
        return True


class Bernoulli(AbstractLikelihood):
    def link_function(self, f: Float[Array, ...]) -> npd.BernoulliProbs:
        r"""The probit link function of the Bernoulli likelihood.

        Args:
            f (Float[Array, "..."]): Function values.

        Returns:
            npd.Bernoulli: The likelihood function.
        """
        return npd.Bernoulli(probs=inv_probit(f))

    def predict(
        self, dist: tp.Union[npd.MultivariateNormal, GaussianDistribution]
    ) -> npd.BernoulliProbs:
        r"""Evaluate the pointwise predictive distribution.

        Evaluate the pointwise predictive distribution, given a Gaussian
        process posterior and likelihood parameters.

        Args:
            dist ([npd.MultivariateNormal, GaussianDistribution].): The Gaussian
                process posterior, evaluated at a finite set of test points.

        Returns:
            npd.Bernoulli: The pointwise predictive distribution.
        """
        variance = jnp.diag(dist.covariance_matrix)
        mean = dist.mean.ravel()
        return self.link_function(mean / jnp.sqrt(1.0 + variance))


class Poisson(AbstractLikelihood):
    def link_function(self, f: Float[Array, ...]) -> npd.Poisson:
        r"""The link function of the Poisson likelihood.

        Args:
            f (Float[Array, "..."]): Function values.

        Returns:
            npd.Poisson: The likelihood function.
        """
        return npd.Poisson(rate=jnp.exp(f))

    def predict(
        self, dist: tp.Union[npd.MultivariateNormal, GaussianDistribution]
    ) -> npd.Poisson:
        r"""Evaluate the pointwise predictive distribution.

        Evaluate the pointwise predictive distribution, given a Gaussian
        process posterior and likelihood parameters.

        Args:
            dist (tp.Union[npd.MultivariateNormal, GaussianDistribution]): The Gaussian
                process posterior, evaluated at a finite set of test points.

        Returns:
            npd.Poisson: The pointwise predictive distribution.
        """
        return self.link_function(dist.mean)


def inv_probit(x: Float[Array, " *N"]) -> Float[Array, " *N"]:
    r"""Compute the inverse probit function.

    Args:
        x (Float[Array, "*N"]): A vector of values.

    Returns
    -------
    Float[Array, "*N"]
        The inverse probit of the input vector.
    """
    jitter = 1e-3  # To ensure output is in interval (0, 1).
    return 0.5 * (1.0 + jsp.special.erf(x / jnp.sqrt(2.0))) * (1 - 2 * jitter) + jitter


NonGaussian = tp.Union[Poisson, Bernoulli]

__all__ = [
    "AbstractHeteroscedasticLikelihood",
    "AbstractLikelihood",
    "AbstractNoiseTransform",
    "Bernoulli",
    "Gaussian",
    "HeteroscedasticGaussian",
    "LogNormalTransform",
    "NoiseMoments",
    "NonGaussian",
    "Poisson",
    "SoftplusTransform",
    "inv_probit",
]



import math
import typing as tp

from flax import nnx
import jax
from jax.experimental import checkify
import jax.numpy as jnp
import jax.tree_util as jtu
from jax.typing import ArrayLike
import numpyro.distributions as dist
import numpyro.distributions.transforms as npt

T = tp.TypeVar("T", bound=ArrayLike | list[float])
ParameterTag = str


class FillTriangularTransform(npt.Transform):
    """
    Transform that maps a vector of length n(n+1)/2 to an n x n lower triangular matrix.
    The ordering is assumed to be:
       (0,0), (1,0), (1,1), (2,0), (2,1), (2,2), ..., (n-1, n-1)
    """

    # Note: The base class provides `inv` through _InverseTransform wrapping _inverse.

    def __call__(self, x):
        """
        Forward transformation.

        Parameters
        ----------
        x : array_like, shape (..., L)
            Input vector with L = n(n+1)/2 for some integer n.

        Returns
        -------
        y : array_like, shape (..., n, n)
            Lower-triangular matrix (with zeros in the upper triangle) filled in
            row-major order (i.e. [ (0,0), (1,0), (1,1), ... ]).
        """
        L = x.shape[-1]
        # Use static (Python) math.sqrt to compute n. This avoids tracer issues.
        n = int((-1 + math.sqrt(1 + 8 * L)) // 2)
        if n * (n + 1) // 2 != L:
            raise ValueError("Last dimension must equal n(n+1)/2 for some integer n.")

        def fill_single(vec):
            out = jnp.zeros((n, n), dtype=vec.dtype)
            row, col = jnp.tril_indices(n)
            return out.at[row, col].set(vec)

        if x.ndim == 1:
            return fill_single(x)
        else:
            batch_shape = x.shape[:-1]
            flat_x = x.reshape((-1, L))
            out = jax.vmap(fill_single)(flat_x)
            return out.reshape((*batch_shape, n, n))

    def _inverse(self, y):
        """
        Inverse transformation.

        Parameters
        ----------
        y : array_like, shape (..., n, n)
            Lower triangular matrix.

        Returns
        -------
        x : array_like, shape (..., n(n+1)/2)
            The vector containing the elements from the lower-triangular portion of y.
        """
        if y.ndim < 2:
            raise ValueError("Input to inverse must be at least two-dimensional.")
        n = y.shape[-1]
        if y.shape[-2] != n:
            raise ValueError(f"Input matrix must be square; got shape {y.shape[-2:]}")

        row, col = jnp.tril_indices(n)

        def inv_single(mat):
            return mat[row, col]

        if y.ndim == 2:
            return inv_single(y)
        else:
            batch_shape = y.shape[:-2]
            flat_y = y.reshape((-1, n, n))
            out = jax.vmap(inv_single)(flat_y)
            return out.reshape((*batch_shape, n * (n + 1) // 2))

    def log_abs_det_jacobian(self, x, y, intermediates=None):
        # Since the transform simply reorders the vector into a matrix, the Jacobian determinant is 1.
        return jnp.zeros(x.shape[:-1])

    @property
    def sign(self):
        # The reordering transformation has a positive derivative everywhere.
        return 1.0

    # Implement tree_flatten and tree_unflatten because base Transform expects them.
    def tree_flatten(self):
        # This transform is stateless.
        return (), {}

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls()


def transform(
    params: nnx.State,
    params_bijection: dict[str, npt.Transform],
    inverse: bool = False,
) -> nnx.State:
    r"""Transforms parameters using a bijector.

    Example:
        >>> from gpjax.parameters import PositiveReal, transform
        >>> import jax.numpy as jnp
        >>> import numpyro.distributions.transforms as npt
        >>> from flax import nnx
        >>> params = nnx.State(
        ...     {
        ...         "a": PositiveReal(jnp.array([1.0])),
        ...         "b": PositiveReal(jnp.array([2.0])),
        ...     }
        ... )
        >>> params_bijection = {'positive': npt.SoftplusTransform()}
        >>> transformed_params = transform(params, params_bijection)
        >>> print(transformed_params["a"][...])
        [1.3132617]

    Args:
        params: A nnx.State object containing parameters to be transformed.
        params_bijection: A dictionary mapping parameter types to bijectors.
        inverse: Whether to apply the inverse transformation.

    Returns:
        State: A new nnx.State object containing the transformed parameters.
    """

    def _inner(param):
        bijector = params_bijection.get(param.tag, npt.IdentityTransform())
        if inverse:
            transformed_value = bijector.inv(param[...])
        else:
            transformed_value = bijector(param[...])

        param = param.replace(transformed_value)
        return param

    gp_params, *other_params = nnx.split_state(params, Parameter, ...)

    # Transform each parameter in the state
    transformed_gp_params: nnx.State = jtu.tree_map(
        lambda x: _inner(x) if isinstance(x, Parameter) else x,
        gp_params,
        is_leaf=lambda x: isinstance(x, Parameter),
    )
    return nnx.merge_state(transformed_gp_params, *other_params)


class Parameter(nnx.Variable[T]):
    """Parameter base class.

    All trainable parameters in GPJax should inherit from this class.

    """

    def __init__(
        self,
        value: T,
        tag: ParameterTag,
        prior: dist.Distribution | None = None,
        **kwargs,
    ):
        _check_is_arraylike(value)

        super().__init__(value=jnp.asarray(value), **kwargs)

        # nnx.Variable metadata must be set via set_metadata (direct setattr is disallowed).
        self.set_metadata(
            tag=tag,
            numpyro_properties={"prior": prior} if prior is not None else {},
        )

    @property
    def tag(self) -> ParameterTag:
        """Return the parameter's constraint tag."""
        return self.get_metadata("tag", "real")


class NonNegativeReal(Parameter[T]):
    """Parameter that is non-negative."""

    def __init__(self, value: T, tag: ParameterTag = "non_negative", **kwargs):
        super().__init__(value=value, tag=tag, **kwargs)
        _safe_assert(_check_is_non_negative, self[...])


class PositiveReal(Parameter[T]):
    """Parameter that is strictly positive."""

    def __init__(self, value: T, tag: ParameterTag = "positive", **kwargs):
        super().__init__(value=value, tag=tag, **kwargs)
        _safe_assert(_check_is_positive, self[...])


class Real(Parameter[T]):
    """Parameter that can take any real value."""

    def __init__(self, value: T, tag: ParameterTag = "real", **kwargs):
        super().__init__(value, tag, **kwargs)


class SigmoidBounded(Parameter[T]):
    """Parameter that is bounded between 0 and 1."""

    def __init__(self, value: T, tag: ParameterTag = "sigmoid", **kwargs):
        super().__init__(value=value, tag=tag, **kwargs)

        # Only perform validation in non-JIT contexts
        if (
            not isinstance(value, jnp.ndarray)
            or getattr(value, "aval", None) is not None
        ):
            _safe_assert(
                _check_in_bounds,
                self[...],
                low=jnp.array(0.0),
                high=jnp.array(1.0),
            )


class LowerTriangular(Parameter[T]):
    """Parameter that is a lower triangular matrix."""

    def __init__(self, value: T, tag: ParameterTag = "lower_triangular", **kwargs):
        super().__init__(value=value, tag=tag, **kwargs)

        # Only perform validation in non-JIT contexts
        if (
            not isinstance(value, jnp.ndarray)
            or getattr(value, "aval", None) is not None
        ):
            _safe_assert(_check_is_square, self[...])
            _safe_assert(_check_is_lower_triangular, self[...])


DEFAULT_BIJECTION = {
    "positive": npt.SoftplusTransform(),
    "non_negative": npt.SoftplusTransform(),
    "real": npt.IdentityTransform(),
    "sigmoid": npt.SigmoidTransform(),
    "lower_triangular": FillTriangularTransform(),
}


def _check_is_arraylike(value: T) -> None:
    """Check if a value is array-like.

    Args:
        value: The value to check.

    Raises:
        TypeError: If the value is not array-like.
    """
    if not isinstance(value, (jax.Array, ArrayLike, list)):
        raise TypeError(
            f"Expected parameter value to be an array-like type. Got {value}."
        )


@checkify.checkify
def _check_is_non_negative(value):
    checkify.check(
        jnp.all(value >= 0), "value needs to be non-negative, got {value}", value=value
    )


@checkify.checkify
def _check_is_positive(value):
    checkify.check(
        jnp.all(value > 0), "value needs to be positive, got {value}", value=value
    )


@checkify.checkify
def _check_is_square(value: T) -> None:
    """Check if a value is a square matrix.

    Args:
        value: The value to check.

    Raises:
        ValueError: If the value is not a square matrix.
    """
    checkify.check(
        value.shape[0] == value.shape[1],
        "value needs to be a square matrix, got {value}",
        value=value,
    )


@checkify.checkify
def _check_is_lower_triangular(value: T) -> None:
    """Check if a value is a lower triangular matrix.

    Args:
        value: The value to check.

    Raises:
        ValueError: If the value is not a lower triangular matrix.
    """
    checkify.check(
        jnp.all(jnp.tril(value) == value),
        "value needs to be a lower triangular matrix, got {value}",
        value=value,
    )


@checkify.checkify
def _check_in_bounds(value: T, low: T, high: T) -> None:
    """Check if a value is bounded between low and high.

    Args:
        value: The value to check.
        low: The lower bound.
        high: The upper bound.

    Raises:
        ValueError: If any element of value is outside the bounds.
    """
    checkify.check(
        jnp.all((value >= low) & (value <= high)),
        "value needs to be bounded between {low} and {high}, got {value}",
        value=value,
        low=low,
        high=high,
    )


def _safe_assert(fn: tp.Callable[[tp.Any], None], value: T, **kwargs) -> None:
    error, _ = fn(value, **kwargs)
    checkify.check_error(error)



"""Linear algebra operations for GPJax LinearOperators."""

from jax import Array
import jax.numpy as jnp
import jax.scipy as jsp
from jaxtyping import Float

from gpjax.linalg.operators import (
    BlockDiag,
    Dense,
    Diagonal,
    Identity,
    Kronecker,
    LinearOperator,
    Triangular,
)
from gpjax.typing import ScalarFloat


def lower_cholesky(A: LinearOperator) -> LinearOperator:
    """Compute the lower Cholesky decomposition of a positive semi-definite operator.

    This function dispatches on the type of the input LinearOperator to provide
    efficient implementations for different operator structures.

    Args:
        A: A positive semi-definite LinearOperator.

    Returns:
        The lower triangular Cholesky factor L such that A = L @ L.T.
    """

    def _handle_identity(A):
        return A

    def _handle_diagonal(A):
        return Diagonal(jnp.sqrt(A.diagonal))

    def _handle_triangular(A):
        if A.lower:
            return A
        return Triangular(jnp.linalg.cholesky(A.to_dense()), lower=True)

    def _handle_kronecker(A):
        cholesky_ops = [lower_cholesky(op) for op in A.operators]
        return Kronecker(cholesky_ops)

    def _handle_blockdiag(A):
        cholesky_ops = [lower_cholesky(op) for op in A.operators]
        return BlockDiag(cholesky_ops, multiplicities=A.multiplicities)

    def _handle_dense(A):
        return Triangular(jnp.linalg.cholesky(A.array), lower=True)

    def _handle_default(A):
        return Triangular(jnp.linalg.cholesky(A.to_dense()), lower=True)

    dispatch_table = {
        Identity: _handle_identity,
        Diagonal: _handle_diagonal,
        Triangular: _handle_triangular,
        Kronecker: _handle_kronecker,
        BlockDiag: _handle_blockdiag,
        Dense: _handle_dense,
    }

    handler = dispatch_table.get(type(A), _handle_default)
    return handler(A)


def solve(
    A: LinearOperator,
    b: Float[Array, " N"] | Float[Array, " N M"],
) -> Float[Array, " N"] | Float[Array, " N M"]:
    """Solve the linear system A @ x = b for x.

    This function dispatches on the type of the input LinearOperator to provide
    efficient implementations for different operator structures.

    Args:
        A: A LinearOperator representing the matrix A.
        b: The right-hand side vector or matrix.

    Returns:
        The solution x to the linear system.
    """
    # Handle different shapes of b
    if b.ndim == 1:
        b = b[:, None]
        squeeze_output = True
    else:
        squeeze_output = False

    # Dispatch based on operator type
    if isinstance(A, Identity):
        # Identity matrix: x = b
        result = b

    elif isinstance(A, Diagonal):
        # Diagonal matrix: element-wise division
        result = b / A.diagonal[:, None]

    elif isinstance(A, Triangular):
        # Triangular matrix: use triangular solver
        result = jsp.linalg.solve_triangular(A.array, b, lower=A.lower)

    elif isinstance(A, Dense):
        # Dense matrix: use standard solver
        result = jnp.linalg.solve(A.array, b)

    else:
        # Default: convert to dense and solve
        result = jnp.linalg.solve(A.to_dense(), b)

    if squeeze_output:
        result = result.squeeze(-1)

    return result


def logdet(A: LinearOperator) -> ScalarFloat:
    """Compute the log-determinant of a linear operator.

    This function dispatches on the type of the input LinearOperator to provide
    efficient implementations for different operator structures.

    Args:
        A: A LinearOperator.

    Returns:
        The log-determinant of A.
    """

    def _handle_identity(A):
        return jnp.array(0.0)

    def _handle_diagonal(A):
        return jnp.sum(jnp.log(A.diagonal))

    def _handle_triangular(A):
        diag_elements = jnp.diag(A.array)
        return jnp.sum(jnp.log(diag_elements))

    def _handle_kronecker(A):
        logdet_val = 0.0
        for i, op in enumerate(A.operators):
            op_logdet = logdet(op)
            power = 1
            for j, other_op in enumerate(A.operators):
                if i != j:
                    power *= other_op.shape[0]
            logdet_val += power * op_logdet
        return logdet_val

    def _handle_blockdiag(A):
        logdet_val = 0.0
        for op, mult in zip(A.operators, A.multiplicities, strict=False):
            logdet_val += mult * logdet(op)
        return logdet_val

    def _handle_dense(A):
        _, logdet_val = jnp.linalg.slogdet(A.array)
        return logdet_val

    def _handle_default(A):
        _, logdet_val = jnp.linalg.slogdet(A.to_dense())
        return logdet_val

    dispatch_table = {
        Identity: _handle_identity,
        Diagonal: _handle_diagonal,
        Triangular: _handle_triangular,
        Kronecker: _handle_kronecker,
        BlockDiag: _handle_blockdiag,
        Dense: _handle_dense,
    }

    handler = dispatch_table.get(type(A), _handle_default)
    return handler(A)


def diag(A: LinearOperator) -> Float[Array, " N"]:
    """Extract the diagonal of a linear operator.

    This function dispatches on the type of the input LinearOperator to provide
    efficient implementations for different operator structures.

    Args:
        A: A LinearOperator.

    Returns:
        The diagonal elements of A as a 1D array.
    """

    def _handle_identity(A):
        n = A.shape[0]
        return jnp.ones(n, dtype=A.dtype)

    def _handle_diagonal(A):
        return A.diagonal

    def _handle_triangular(A):
        return jnp.diag(A.array)

    def _handle_kronecker(A):
        result = diag(A.operators[0])
        for op in A.operators[1:]:
            result = jnp.kron(result, diag(op))
        return result

    def _handle_blockdiag(A):
        diags = []
        for op, mult in zip(A.operators, A.multiplicities, strict=False):
            op_diag = diag(op)
            for _ in range(mult):
                diags.append(op_diag)
        return jnp.concatenate(diags)

    def _handle_dense(A):
        return jnp.diag(A.array)

    def _handle_default(A):
        return jnp.diag(A.to_dense())

    dispatch_table = {
        Identity: _handle_identity,
        Diagonal: _handle_diagonal,
        Triangular: _handle_triangular,
        Kronecker: _handle_kronecker,
        BlockDiag: _handle_blockdiag,
        Dense: _handle_dense,
    }

    handler = dispatch_table.get(type(A), _handle_default)
    return handler(A)



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.19.1
#   kernelspec:
#     display_name: .venv
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Heteroscedastic Inference
#
# This notebook shows how to fit a heteroscedastic Gaussian processes (GPs) that
# allows one to perform regression where there exists non-constant, or
# input-dependent, noise.
#
#
# ## Background
# A heteroscedastic GP couples two latent functions:
# - A **signal GP** $f(\cdot)$ for the mean response.
# - A **noise GP** $g(\cdot)$ that maps to a positive variance
#   $\sigma^2(x) = \phi(g(x))$ via a positivity transform $\phi$ (typically
#   ${\rm exp}$ or ${\rm softplus}$). Intuitively, we are introducing a pair of GPs;
# one to model the latent mean, and a second that models the log-noise variance. This
# is in direct contrast a
# [homoscedastic GP](https://docs.jaxgaussianprocesses.com/_examples/regression/)
# where we learn a constant value for the noise.
#
# In the Gaussian case, the observed response follows
# $$y \mid f, g \sim \mathcal{N}(f, \sigma^2(x)).$$
# Variational inference works with independent posteriors $q(f)q(g)$, combining the
# moments of each into an ELBO. For non-Gaussian likelihoods the same structure
# remains; only the expected log-likelihood changes.

# %%
from jax import config
import jax.numpy as jnp
import jax.random as jr
import matplotlib as mpl
import matplotlib.pyplot as plt
import optax as ox

from examples.utils import use_mpl_style
import gpjax as gpx
from gpjax.likelihoods import (
    HeteroscedasticGaussian,
    LogNormalTransform,
    SoftplusTransform,
)
from gpjax.variational_families import (
    HeteroscedasticVariationalFamily,
    VariationalGaussianInit,
)

# Enable Float64 for stable linear algebra.
config.update("jax_enable_x64", True)


use_mpl_style()
key = jr.key(123)
cols = mpl.rcParams["axes.prop_cycle"].by_key()["color"]


# %% [markdown]
# ## Dataset simulation
# We simulate whose mean and noise levels vary with
# the input. We sample inputs $x \sim \mathcal{U}(0, 1)$ and define the
# latent signal to be
# $$f(x) = (x - 0.5)^2 + 0.05;$$
# a smooth bowl-shaped curve. The observation standard deviation is chosen to be
# proportional to the signal,
# $$\sigma(x) = 0.5\,f(x),$$
# which yields the heteroscedastic generative model
# $$y \mid x \sim \mathcal{N}\!\big(f(x), \sigma^2(x)\big).$$
# This construction makes the noise small near the minimum of the bowl and much
# larger in the tails. We also create a dense test grid that we shall use later for
# visualising posterior fits and predictive uncertainty.

# %%
# Create data with input-dependent variance.
key, x_key, noise_key = jr.split(key, 3)
n = 200
x = jr.uniform(x_key, (n, 1), minval=0.0, maxval=1.0)
signal = (x - 0.5) ** 2 + 0.05
noise_scale = 0.5 * signal
noise = noise_scale * jr.normal(noise_key, shape=(n, 1))
y = signal + noise
train = gpx.Dataset(X=x, y=y)

xtest = jnp.linspace(-0.1, 1.1, 200)[:, None]
signal_test = (xtest - 0.5) ** 2 + 0.05
noise_scale_test = 0.5 * signal_test
noise_test = noise_scale_test * jr.normal(noise_key, shape=(200, 1))
ytest = signal_test + noise_test

fig, ax = plt.subplots()
ax.plot(x, y, "o", label="Observations", alpha=0.7, color=cols[0])
ax.plot(xtest, signal_test, label="Signal", alpha=0.7, color=cols[1])
ax.plot(xtest, noise_scale_test, label="Noise scale", alpha=0.7, color=cols[2])
ax.set_xlabel("$x$")
ax.set_ylabel("$y$")
ax.legend(loc="upper left")

# %% [markdown]
# For a homoscedastic baseline, compare this figure with the
# [Gaussian process regression notebook](https://docs.jaxgaussianprocesses.com/_examples/regression/)
# (`examples/regression.py`), where a single latent GP is paired with constant
# observation noise.

# %% [markdown]
# ## Prior specification
# We place independent Gaussian process priors on the signal and noise processes:
# $$f \sim \mathcal{GP}\big(0, k_f\big), \qquad g \sim \mathcal{GP}\big(0, k_g\big),$$
# where $k_f$ and $k_g$ are stationary squared-exponential kernels with unit
# variance and lengthscale of one. The noise process $g$ is mapped to the variance
# via the logarithmic transform in `LogNormalTransform`, giving
# $\sigma^2(x) = \exp\big(g(x)\big)$. The joint prior over $(f, g)$ combines with
# the heteroscedastic Gaussian likelihood,
# $$p(\mathbf{y} \mid f, g) = \prod_{i=1}^n
# \mathcal{N}\!\big(y_i \mid f(x_i), \exp(g(x_i))\big),$$
# to form the posterior target that we shall approximate variationally. The product
# syntax `signal_prior * likelihood` used below constructs this augmented GP model.

# %%
# Signal and noise priors.
signal_prior = gpx.gps.Prior(
    mean_function=gpx.mean_functions.Zero(),
    kernel=gpx.kernels.RBF(),
)
noise_prior = gpx.gps.Prior(
    mean_function=gpx.mean_functions.Zero(),
    kernel=gpx.kernels.RBF(),
)
likelihood = HeteroscedasticGaussian(
    num_datapoints=train.n,
    noise_prior=noise_prior,
    noise_transform=LogNormalTransform(),
)
posterior = signal_prior * likelihood

# Variational family over both processes.
z = jnp.linspace(-3.2, 3.2, 25)[:, None]
q = HeteroscedasticVariationalFamily(
    posterior=posterior,
    inducing_inputs=z,
    inducing_inputs_g=z,
)

# %% [markdown]
# The variational family introduces inducing variables for both latent functions,
# located at the set $Z = \{z_m\}_{m=1}^M$. These inducing variables summarise the
# infinite-dimensional GP priors in terms of multivariate Gaussian parameters.
# Optimising the evidence lower bound (ELBO) corresponds to adjusting the means and
# covariances of the variational posteriors $q(f)$ and $q(g)$ so that they best
# explain the observed data whilst remaining close to the prior. For a deeper look at
# these constructions in the homoscedastic setting, refer to the
# [Sparse Gaussian Process Regression](https://docs.jaxgaussianprocesses.com/_examples/collapsed_vi/)
# (`examples/collapsed_vi.py`) and
# [Sparse Stochastic Variational Inference](https://docs.jaxgaussianprocesses.com/_examples/uncollapsed_vi/)
# (`examples/uncollapsed_vi.py`) notebooks.

# %% [markdown]
# ### Optimisation
# With the model specified, we minimise the negative ELBO,
# $$\mathcal{L} = \mathbb{E}_{q(f)q(g)}\!\big[\log p(\mathbf{y}\mid f, g)\big]
# - \mathrm{KL}\!\left[q(f) \,\|\, p(f)\right]
# - \mathrm{KL}\!\left[q(g) \,\|\, p(g)\right],$$
# using the Adam optimiser. GPJax automatically selects the tight bound of
# Lázaro-Gredilla & Titsias (2011) when the likelihood is Gaussian, yielding an
# analytically tractable expectation over the latent noise process. The resulting
# optimisation iteratively updates the inducing posteriors for both latent GPs.

# %%
# Optimise the heteroscedastic ELBO (selects tighter bound).
objective = lambda model, data: -gpx.objectives.heteroscedastic_elbo(model, data)
optimiser = ox.adam(1e-2)
q_trained, history = gpx.fit(
    model=q,
    objective=objective,
    train_data=train,
    optim=optimiser,
    num_iters=10000,
    verbose=False,
)

loss_trace = jnp.asarray(history)
print(f"Final regression ELBO: {-loss_trace[-1]:.3f}")

# %% [markdown]
# ## Prediction
# After training we obtain posterior marginals for both latent functions. To make a
# prediction we evaluate two quantities:
# 1. The latent posterior over $f$ (mean and variance), which reflects uncertainty
#    in the latent function **prior** to observing noise.
# 2. The marginal predictive over observations, which integrates out both $f$ and
#    $g$ to provide predictive intervals for future noisy measurements.
# The helper method `likelihood.predict` performs the second integration for us.

# %%
# Predict on a dense grid.
xtest = jnp.linspace(-0.1, 1.1, 200)[:, None]
mf, vf, mg, vg = q_trained.predict(xtest)

signal_pred, noise_pred = q_trained.predict_latents(xtest)
predictive = likelihood.predict(signal_pred, noise_pred)

fig, ax = plt.subplots()
ax.plot(train.X, train.y, "o", label="Observations", alpha=0.5)
ax.plot(xtest, mf, color="C0", label="Posterior mean")
ax.fill_between(
    xtest.squeeze(),
    (mf.squeeze() - 2 * jnp.sqrt(vf.squeeze())).squeeze(),
    (mf.squeeze() + 2 * jnp.sqrt(vf.squeeze())).squeeze(),
    color="C0",
    alpha=0.15,
    label="±2 std (latent)",
)
ax.fill_between(
    xtest.squeeze(),
    predictive.mean - 2 * jnp.sqrt(jnp.diag(predictive.covariance_matrix)),
    predictive.mean + 2 * jnp.sqrt(jnp.diag(predictive.covariance_matrix)),
    color="C1",
    alpha=0.15,
    label="±2 std (observed)",
)
ax.set_xlabel("$x$")
ax.set_ylabel("$y$")
ax.legend(loc="upper left")
ax.set_title("Heteroscedastic regression")

# %% [markdown]
# The latent intervals quantify epistemic uncertainty about $f$, whereas the broader
# observed band adds the aleatoric noise predicted by $g$. The widening of the orange
# band in the right half matches the ground-truth construction of the dataset.

# %% [markdown]
# ## Sparse Heteroscedastic Regression
#
# We now demonstrate how the aforementioned heteroscedastic approach can be extended
# into sparse scenarios, thus offering more favourable scalability as the size of our
# dataset grows. To achieve this we defined inducing points for both the signal and
# noise processes. Decoupling these grids allows us to focus modelling
# capacity where each latent function varies the most. The synthetic dataset below
# contains a smooth sinusoidal signal but exhibits a sharply peaked noise shock,
# mimicking the situation where certain regions of the input space are far noisier
# than others.

# %%
# Generate data
key, x_key, noise_key = jr.split(key, 3)
n = 300
x = jr.uniform(x_key, (n, 1), minval=-2.0, maxval=2.0)
signal = jnp.sin(2.0 * x)
# Gaussian bump of noise
noise_std = 0.1 + 0.5 * jnp.exp(-0.5 * ((x - 0.5) / 0.4) ** 2)
y = signal + noise_std * jr.normal(noise_key, shape=(n, 1))
data_adv = gpx.Dataset(X=x, y=y)

# %% [markdown]
# ### Model components
# We again adopt RBF priors for both processes but now apply a `SoftplusTransform`
# to the noise GP. This alternative map enforces positivity whilst avoiding the
# heavier tails induced by the log-normal transform. The `HeteroscedasticGaussian`
# likelihood seamlessly accepts the new transform.

# %%
# Define model components
mean_prior = gpx.gps.Prior(
    mean_function=gpx.mean_functions.Zero(),
    kernel=gpx.kernels.RBF(),
)
noise_prior_adv = gpx.gps.Prior(
    mean_function=gpx.mean_functions.Zero(),
    kernel=gpx.kernels.RBF(),
)
likelihood_adv = HeteroscedasticGaussian(
    num_datapoints=data_adv.n,
    noise_prior=noise_prior_adv,
    noise_transform=SoftplusTransform(),
)
posterior_adv = mean_prior * likelihood_adv

# %%
# Configure variational family
# The signal requires a richer inducing set to capture its oscillations, whereas the
# noise process can be summarised with fewer points because the burst is localised.
z_signal = jnp.linspace(-2.0, 2.0, 30)[:, None]
z_noise = jnp.linspace(-2.0, 2.0, 20)[:, None]

# Use VariationalGaussianInit to pass specific configurations
q_init_f = VariationalGaussianInit(inducing_inputs=z_signal)
q_init_g = VariationalGaussianInit(inducing_inputs=z_noise)

q_sparse = HeteroscedasticVariationalFamily(
    posterior=posterior_adv,
    signal_init=q_init_f,
    noise_init=q_init_g,
)

# %% [markdown]
# The initialisation objects `VariationalGaussianInit` allow us to prescribe
# different inducing grids and initial covariance structures for $f$ and $g$. This
# flexibility is invaluable when working with large datasets where the latent
# functions have markedly different smoothness properties.

# %%
# Optimize
objective_adv = lambda model, data: -gpx.objectives.heteroscedastic_elbo(model, data)
optimiser_adv = ox.adam(1e-2)
q_sparse_trained, _ = gpx.fit(
    model=q_sparse,
    objective=objective_adv,
    train_data=data_adv,
    optim=optimiser_adv,
    num_iters=10000,
    verbose=False,
)

# %%
# Plotting
xtest = jnp.linspace(-2.2, 2.2, 300)[:, None]
pred = q_sparse_trained.predict(xtest)

# Unpack the named tuple
mf = pred.mean_f
vf = pred.variance_f
mg = pred.mean_g
vg = pred.variance_g

# Calculate total predictive variance
# The likelihood expects the *latent* noise distribution to compute the predictive
# but here we can just use the transformed expected variance for plotting.
# For accurate predictive intervals, we should use likelihood.predict.
signal_dist, noise_dist = q_sparse_trained.predict_latents(xtest)
predictive_dist = likelihood_adv.predict(signal_dist, noise_dist)
predictive_mean = predictive_dist.mean
predictive_std = jnp.sqrt(jnp.diag(predictive_dist.covariance_matrix))

fig, ax = plt.subplots(figsize=(6, 2.5))
ax.plot(x, y, "x", color="black", alpha=0.5, label="Data")

# Plot total uncertainty (signal + noise)
ax.plot(xtest, predictive_mean, "--", color=cols[1], linewidth=2)
ax.fill_between(
    xtest.squeeze(),
    predictive_mean - predictive_std,
    predictive_mean + predictive_std,
    color=cols[1],
    alpha=0.3,
    label="One std. dev.",
)
ax.plot(
    xtest.squeeze(),
    predictive_mean - predictive_std,
    "--",
    color=cols[1],
    alpha=0.5,
    linewidth=0.75,
)
ax.plot(
    xtest.squeeze(),
    predictive_mean + predictive_std,
    "--",
    color=cols[1],
    alpha=0.5,
    linewidth=0.75,
)
ax.fill_between(
    xtest.squeeze(),
    predictive_mean - 2 * predictive_std,
    predictive_mean + 2 * predictive_std,
    color=cols[1],
    alpha=0.1,
    label="Two std. dev.",
)
ax.plot(
    xtest.squeeze(),
    predictive_mean - 2 * predictive_std,
    "--",
    color=cols[1],
    alpha=0.5,
    linewidth=0.75,
)
ax.plot(
    xtest.squeeze(),
    predictive_mean + 2 * predictive_std,
    "--",
    color=cols[1],
    alpha=0.5,
    linewidth=0.75,
)

ax.set_title("Sparse Heteroscedastic Regression")
ax.legend(loc="best", fontsize="small")
ax.set_xlabel("$x$")
ax.set_ylabel("$y$")

# %% [markdown]
# ## Takeaways
# - The heteroscedastic GP model couples two latent GPs, enabling separate control of
#   epistemic and aleatoric uncertainties.
# - We support multiple positivity transforms for the noise process; the choice
#   affects the implied variance tails and should reflect prior beliefs.
# - Inducing points for the signal and noise processes can be tuned independently to
#   balance computational budget against the local complexity of each function.
# - The ELBO implementation automatically selects the tightest analytical bound
#   available, streamlining heteroscedastic inference workflows.

# %% [markdown]
# ## System configuration

# %%
# %reload_ext watermark
# %watermark -n -u -v -iv -w -a 'Thomas Pinder'
